//Using SDL and standard IO
#include <SDL.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ted_maps.h"
#include "cmd_lib.h"

#define	MAXTOTALSEGS	8192/4
#define	MAXNODES		2048
#define	BLOCKSCALE		8

#define	PUSHWALL		98
#define	SLOTTILE		99
#define	AREATILES		108
#define	DOORTILE		90
#define	LASTDOORTILE	97

#define SCALE           3

#define BOTTOM_WALL_SOLID    1
#define TOP_WALL_SOLID       2
#define BOTH_WALLS_SOLID     3

#define	DIR_SEGFLAG		     0x80
#define	DIR_LASTSEGFLAG	     0x40

typedef enum { or_horizontal, or_vertical } orientation_t;

// the direction is from point 1 (view left) to point 2 (view right)
typedef enum { di_north, di_east, di_south, di_west } dir_t;

typedef struct
{
	unsigned char		tilemap[64][64];
	unsigned char		areasoundnum[64];
	short	numspawn;
	short	spawnlistofs;
	short	numnodes;
	short	nodelistofs;
	unsigned char		data[0x8000];		// nodes, and spawn list
} loadmap_t;

typedef struct segment_s
{
	struct segment_s* next;
	orientation_t	orientation;
	dir_t			dir;
	int			coordinate;
	int			min, max;
	int			texture;
} segment_t;

typedef struct
{
	int	x; 
	int y;
	int area;
	int tile;
	int psegnum;
} pwall_t;

typedef struct
{
	segment_t* segs;		        // if non NULL, its a terminal node

	orientation_t	orientation;	// only valid if segs is NULL
	int			coordinate;
	int			min, max;
	int			frontnode, backnode;
} bspnode_t;

typedef struct
{
	unsigned char plane;
	unsigned char dir;
	unsigned short children[2];
} savenode_t;

typedef struct
{
	unsigned char 		plane;			// in half tiles
	unsigned char 		dir;
	unsigned char 		min, max;	    // in half tiles
	unsigned char 		texture;
	unsigned char 		area;
} saveseg_t;


//Screen dimension constants
const int SCREEN_WIDTH = BLOCKSCALE * 64;
const int SCREEN_HEIGHT = BLOCKSCALE * 64;

//The MAP window we'll be rendering to
SDL_Window* gMapWindow = NULL;

//The map window renderer
SDL_Renderer* gMapRenderer = NULL;

//The seg window we'll be rendering to
SDL_Window* gSegWindow = NULL;

//The seg window renderer
SDL_Renderer* gSegRenderer = NULL;

// map for loading data into
static loadmap_t	map;

static unsigned short back[64][64];
static unsigned short front[64][64];

static segment_t segs[MAXTOTALSEGS], * newseg_p;
static bspnode_t nodes[MAXNODES], * node_p;
static savenode_t* savenodes, * savenode_p;

static pwall_t	pwalls[128];
static int numpwalls;

static int doornum[64][64];
static int numdoors;

static unsigned char areamap[64][64];

static int depth = 0;

static bool init();
static void close();
static void SetupMap(int num);
static void DrawMap(void);
static void DrawTile(int x, int y, int color);
static void RemovePWalls(void);
static void AddSeg(dir_t dir, float coordinate, int min, int max, int texture);
static void DrawSeg(SDL_Renderer* renderer, segment_t* seg);
static void FindDoors(void);
static void CreateSegments(void);
static void FloodMap(void);
static void FloodFill(int x, int y, int fill);
static void BSPMap(void);
static int BSPList(segment_t* list, float xl, float yl, float xh, float yh);
static void CountSegSplits(segment_t* list, orientation_t 	splitdir, int splitpoint, int* frontcnt, int* backcnt);
static void SplitSegs(segment_t* list, orientation_t splitdir, int splitpoint, segment_t** frontlist, segment_t** backlist);
static void DrawCircle(int32_t centreX, int32_t centreY, int32_t radius);
static void SaveMap(int area);
static void OutputNode(int nodenum);
static void OutputSegs(segment_t* seg);

/*
================
=
= main
=
================
*/
int main( int argc, char* args[] )
{
	int	i, start, stop;

	if (argc == 2)
	{
		start = stop = atoi(args[1]);
	}
	else if (argc == 3)
	{
		start = atoi(args[1]);
		stop = atoi(args[2]);
	}
	else
	{
		start = 0;
		stop = 0;
	}

	/* Start up SDLand create window */
	if (!init())
	{
		printf("Failed to initialize!\n");
	}
	else
	{
		//Main loop flag
		bool quit = false;

		i = start;

		//Event handler
		SDL_Event e;

		LoadTedHeader("n3d");
		printf("Header OK\n");

		//While application is running
		while (!quit)
		{
			//Handle events on queue
			while (SDL_PollEvent(&e) != 0)
			{
				//User requests quit for map window
				if (e.type == SDL_QUIT)
				{
					quit = true;
				}

				if (e.type == SDL_WINDOWEVENT
					&& e.window.event == SDL_WINDOWEVENT_CLOSE)
				{
					// Handle window close segs window
					if (SDL_GetWindowID(gSegWindow) == e.window.windowID)
					{
						SDL_DestroyRenderer(gSegRenderer);
						SDL_DestroyWindow(gSegWindow);

						gSegWindow = NULL;
						gSegRenderer = NULL;
					}
				}
			}

			for (; i <= stop; i++)
			{
				//Clear screen
				SDL_SetRenderDrawColor(gMapRenderer, 0xFF, 0xFF, 0xFF, 0xFF);
				SDL_RenderClear(gMapRenderer);

				SetupMap(i);
				SDL_SetWindowTitle(gMapWindow, header.name);
				printf("\nmap: %i (%s)\n", i, header.name);
				BSPMap();
				SaveMap(i);
				free(mapplanes[0]);
				free(mapplanes[1]);
				free(mapplanes[2]);
			}
		}
	}

	//Free resources and close SDL
	close();

	return 0;
}

/*
================
=
= SetupMap
=
================
*/
static void SetupMap(int num)
{
	int	x;

	LoadMap(num);

	// loads into back / front
	for (x = 0; x < 4096; x++)
	{
		back[0][x] = mapplanes[0][x];
		front[0][x] = mapplanes[1][x];
	}

	newseg_p = segs;
	DrawMap();
	memset(&map, 0, sizeof(map));
	RemovePWalls();
	SDL_RenderPresent(gMapRenderer);
	FindDoors();
	CreateSegments();
	FloodMap();
}

/*
================
=
= BSPMap
=
================
*/
static void BSPMap(void)
{
	segment_t* check;
	int	startnode;

	/* link the segs into one list */
	for (check = segs; check < newseg_p - 1; check++)
	{
		check->next = check + 1;
	}

	check->next = 0;

	node_p = nodes;

	SDL_RenderSetScale(gMapRenderer, SCALE, SCALE);

	startnode = BSPList(segs, 0, 0, 64, 64);
}

/*
================
=
= SaveMap
=
================
*/
static void SaveMap(int area)
{
	char		name[1024];
	FILE* pHandle;
	int		 numactors, numstatics;
	unsigned char *dat;
	int		i, len, x, y, tile;
	int		numsavenodes;

	numactors = 0;
	numstatics = 0;

	printf("saving map...\n");

	/* spawn list */
	map.numspawn = 0;

	dat = map.data;
	map.spawnlistofs = (short)(dat - (unsigned char *)&map);

	for (y = 0; y < 64; y++)
	{
		for (x = 0; x < 64; x++)
		{
			tile = front[y][x];

			if (tile)
			{
				if (tile == SLOTTILE)
				{
					continue;
				}

				*dat++ = x;
				*dat++ = y;
				*dat++ = tile;

				if(tile == PUSHWALL)
				{	// pushwalls get an extra byte for the tile number
					for (i = 0; i < numpwalls; i++)
						if (pwalls[i].x == x && pwalls[i].y == y)
							break;
					*dat++ = pwalls[i].tile;
				}

				map.numspawn++;

				if (tile >= 108)
				{
					numactors++;

					/* set colour to yellow */
					SDL_SetRenderDrawColor(gMapRenderer, 0xff, 0xff, 0x55, 0xFF);
					SDL_RenderSetScale(gMapRenderer, 1.0f, 1.0f);
					DrawCircle((x * BLOCKSCALE) + 3, (y * BLOCKSCALE) + 3, BLOCKSCALE / 2);
				}
				else if (tile >= 23 && tile <= 54)
				{
					numstatics++;
					
					/* set colour to blue  */
					SDL_SetRenderDrawColor(gMapRenderer, 0x00, 0x00, 0xaa, 0xFF);
					SDL_RenderSetScale(gMapRenderer, 1.0f, 1.0f);
					DrawCircle((x * BLOCKSCALE) + 3, (y * BLOCKSCALE) + 3, BLOCKSCALE / 2);
				}
			}
		}
	}

	/* Update screen */
	SDL_RenderPresent(gMapRenderer);

	printf("spawn: %i (%i actors, %i statics)\n", map.numspawn, numactors, numstatics);

	/* compactand save the segs */
	map.nodelistofs = (short)(dat - (unsigned char *)&map);
	savenode_p = savenodes = (savenode_t*)dat;
	OutputNode(0);
	numsavenodes = (int)(savenode_p - savenodes);
	map.numnodes = (short)numsavenodes;
	printf("nodes: %i \n", numsavenodes);

	/* write it out */
	len = (int)((unsigned char *)savenode_p - (unsigned char *)&map);
	sprintf(name, "map%02i.bin", area);
	pHandle = SafeOpenWrite(name);
	SafeWrite(pHandle, &map, len);
	fclose(pHandle);
	printf("map size: %i\n", len);
}

/*
====================
=
= OutputNode
=
====================
*/
static void OutputNode(int nodenum)
{
	bspnode_t* node;
	savenode_t* snode;

	node = &nodes[nodenum];

	/* cehck for termination node */
	if (node->segs)
	{
		OutputSegs(node->segs);
		return;
	}

	snode = savenode_p++;
	snode->dir = node->orientation;
	snode->plane = node->coordinate >> 7;
	snode->children[0] = (unsigned short)(savenode_p - savenodes);  // OPTIMIZE: this is invariant
	OutputNode(node->frontnode);
	snode->children[1] = (unsigned short)(savenode_p - savenodes);
	OutputNode(node->backnode);
}

/*
===============
=
= OutputSegs
=
= Adds the segments to the save list 
===============
*/
static void OutputSegs(segment_t* seg)
{
	int			x, y;
	saveseg_t* saveseg_p = NULL;

	for (; seg; seg = seg->next)
	{
		saveseg_p = (saveseg_t*)savenode_p;
		savenode_p++;

		saveseg_p->dir = seg->dir | DIR_SEGFLAG;
		saveseg_p->plane = seg->coordinate >> 7;  // cut to half tiles
		saveseg_p->min = seg->min >> 7;		// cut down to half tiles
		saveseg_p->max = seg->max >> 7;	// cut down to half tiles
		saveseg_p->texture = seg->texture;

		// find the area the seg bounds
		switch (seg->dir)
		{	// get coorinates of an open tile bordering the seg
		case di_north:
			x = seg->coordinate + 32;
			y = (seg->min + seg->max) >> 1;
			break;
		case di_south:
			x = seg->coordinate - 32;
			y = (seg->min + seg->max) >> 1;
			break;
		case di_east:
			y = seg->coordinate + 32;
			x = (seg->min + seg->max) >> 1;
			break;
		case di_west:
			y = seg->coordinate - 32;
			x = (seg->min + seg->max) >> 1;
			break;
		}

		saveseg_p->area = map.tilemap[y >> 8][x >> 8];

		if (saveseg_p->area >= 63)
		{
			if (saveseg_p->area >= 0x80 + DOORTILE
				&& saveseg_p->area <= 0x80 + LASTDOORTILE)
			{
				if (saveseg_p->area & 1)
				{	// horizontal door
					if ((y & 255) < 128)
						y -= 256;
					else
						y += 256;
				}
				else
				{	// vertical door
					if ((x & 255) < 128)
						x -= 256;
					else
						x += 256;
				}
				saveseg_p->area = map.tilemap[y >> 8][x >> 8];
			}
			else
			{
				saveseg_p->area = 62;	// pushwall flush side
			}
		}
	}

	saveseg_p->dir |= DIR_LASTSEGFLAG;
}

/*
================
=
= BSPList
=
= Returns the node for the given list
================
*/
static int BSPList(segment_t* list, float xl, float yl, float xh, float yh)
{
	bspnode_t* thisnode_p;
	segment_t* check, * bestcheck, * frontlist, * backlist;
	int		   node;
	int		   totalsegs, bestnew, bestsplit, newsegs, split, front, back;
	float	   floatsplit;

	SDL_Point points[5] = 
	{
	    {((int)xl * BLOCKSCALE) / SCALE, ((int)yl * BLOCKSCALE) / SCALE},
	    {((int)xh * BLOCKSCALE) / SCALE, ((int)yl * BLOCKSCALE) / SCALE},
	    {((int)xh * BLOCKSCALE) / SCALE, ((int)yh * BLOCKSCALE) / SCALE},
	    {((int)xl * BLOCKSCALE) / SCALE, ((int)yh * BLOCKSCALE) / SCALE},
		{((int)xl * BLOCKSCALE) / SCALE, ((int)yl * BLOCKSCALE) / SCALE}
	};

	/* set colour to white */
	SDL_SetRenderDrawColor(gMapRenderer, 0xff, 0xff, 0xff, 0xff);

	/* draw rectangle around segs */
	SDL_RenderDrawLines(gMapRenderer, points, 5);

	/* Update the map screen */
	SDL_RenderPresent(gMapRenderer);

	/* count segs in list */
	totalsegs = 0;

	for (check = list; check; check = check->next)
	{
		totalsegs++;
	}

	/* find the best split point */
	bestsplit = 0;
	bestnew = 0xffff;
	bestcheck = NULL;

	for (check = list; check; check = check->next)
	{
		CountSegSplits(list, check->orientation, check->coordinate, &front, &back);

		if (!front || !back)
		{
			continue;    /* must seperate at least one seg */
		}

		newsegs = front + back - totalsegs;

		split = front > back ? back : front;

		if (newsegs <= bestnew && split > bestsplit)
		{
			bestnew = newsegs;
			bestsplit = split;
			bestcheck = check;
		}
	}

	/* split it */
	thisnode_p = node_p;
	node = (int)(node_p - nodes);
	node_p++;

	if ((unsigned)bestnew == 0xffff)
	{	/* remaining segs are convex, no need to split */
		thisnode_p->segs = list;
		return node;
	}

	CountSegSplits(list, bestcheck->orientation, bestcheck->coordinate, &front, &back);
	printf("front: %i  back: %i  new: %i  split: %i\n", front, back, bestnew, bestsplit);

#if 1
	//Clear segs screen
	SDL_SetRenderDrawColor(gSegRenderer, 0x00, 0x00, 0x00, 0xFF);        //black
	SDL_RenderClear(gSegRenderer);

	SDL_SetRenderDrawBlendMode(gSegRenderer, SDL_BLENDMODE_BLEND);
	SDL_SetRenderDrawColor(gSegRenderer, 0xff, 0xff, 0xff, 0x40);        //white
	SDL_RenderSetScale(gSegRenderer, SCALE, SCALE);
	SDL_RenderDrawLines(gSegRenderer, points, 5);
	SDL_RenderPresent(gSegRenderer);
	SDL_SetRenderDrawBlendMode(gSegRenderer, SDL_BLENDMODE_NONE);

	/* display the segs */
	if (depth)
	{
		SDL_SetRenderDrawColor(gSegRenderer, 0xff, 0x00, 0x00, 0xFF);    // Red
	}
	else
	{
		SDL_SetRenderDrawColor(gSegRenderer, 0x00, 0xff, 0x00, 0xFF);    // Green
	}

	depth ^= 1;

	for (check = list; check; check = check->next)
	{
		DrawSeg(gSegRenderer, check);
	}

	/* draw that actual line split seg */
	SDL_SetRenderDrawColor(gSegRenderer, 0xff, 0x55, 0xff, 0xFF);    // magenta
	DrawSeg(gSegRenderer, bestcheck);

	/* Update the segs screen */
	SDL_RenderPresent(gSegRenderer);
#endif

	/* slow the loop down so the user can see the lines being drawn */
	SDL_Delay(200);

	thisnode_p->segs = NULL;
	thisnode_p->orientation = bestcheck->orientation;
	thisnode_p->coordinate = bestcheck->coordinate;

	if (thisnode_p->orientation == or_vertical)
	{
		thisnode_p->min = (int)(yl * 2);
		thisnode_p->max = (int)(yh * 2);
	}
	else
	{
		thisnode_p->min = (int)(xl * 2);
		thisnode_p->max = (int)(xh * 2);
	}

	SplitSegs(list, bestcheck->orientation, bestcheck->coordinate, &frontlist, &backlist);

	floatsplit = (float)bestcheck->coordinate / 256;
	
	if (bestcheck->orientation == or_horizontal)
	{
		thisnode_p->frontnode = BSPList(frontlist, xl, floatsplit, xh, yh);
		thisnode_p->backnode = BSPList(backlist, xl, yl, xh, floatsplit);
	}
	else
	{
		thisnode_p->frontnode = BSPList(frontlist, floatsplit, yl, xh, yh);
		thisnode_p->backnode = BSPList(backlist, xl, yl, floatsplit, yh);
	}

	return node;
}

/*
=====================
=
= CountSegSplits
=
= front side is towards more positive coordinates
=====================
*/
static void CountSegSplits(segment_t* list, orientation_t 	splitdir, int splitpoint, int* frontcnt, int* backcnt)
{
	segment_t* seg_p;
	int	backcount, frontcount;

	frontcount = backcount = 0;

	for (seg_p = list; seg_p; seg_p = seg_p->next)
	{
		if (seg_p->orientation == splitdir)
		{
			if (seg_p->coordinate < splitpoint)
			{
				backcount++;
			}
			else if (seg_p->coordinate > splitpoint)
			{
				frontcount++;
			}
			else
			{
				if (seg_p->dir == di_north || seg_p->dir == di_east)
				{
					frontcount++;
				}
				else
				{
					backcount++;
				}
			}
		}
		else
		{
			if (seg_p->max <= splitpoint)
			{
				backcount++;
			}
			else if (seg_p->min >= splitpoint)
			{
				frontcount++;
			}
			else
			{
				backcount++;
				frontcount++;		// split the segment
			}
		}
	}

	*frontcnt = frontcount;
	*backcnt = backcount;
}

/*
=====================
=
= SplitSegs
=
= front side is towards more positive coordinates
=====================
*/
static void SplitSegs(segment_t* list, orientation_t splitdir, int splitpoint, segment_t** frontlist, segment_t** backlist)
{
	segment_t* next, * seg_p, * fronthead, * backhead;

	fronthead = backhead = 0;

	for (seg_p = list; seg_p; seg_p = next)
	{
		next = seg_p->next;

		/* if same line direction only check splitpoint */
		if (seg_p->orientation == splitdir)
		{
			if (seg_p->coordinate < splitpoint)    /* is line seg behind the split line */
			{
				seg_p->next = backhead;            /* add to back link list */
				backhead = seg_p;
			}
			else if (seg_p->coordinate > splitpoint) /* is line seg in front of split line */
			{
				seg_p->next = fronthead;           /* add to front link list */
				fronthead = seg_p;
			}
			else    /* split line is on the same coordinate as this line */
			{
				if (seg_p->dir == di_north || seg_p->dir == di_east)   /* if line is running east or north seg is in front of split line */
				{
					seg_p->next = fronthead;       /* add to front link list */
					fronthead = seg_p;
				}
				else                               /* else seg is behind the split line */
				{
					seg_p->next = backhead;        /* add to back link list */
					backhead = seg_p;
				}
			}
		}
		else   /* split line is opposite orientation to split line (need to check min and max coordinate) */
		{
			if (seg_p->max <= splitpoint)    /* is line seg completely behind the split line */
			{
				seg_p->next = backhead;      /* add to back link list */
				backhead = seg_p;
			}
			else if (seg_p->min >= splitpoint) /* is line seg completely in front of split line */
			{
				seg_p->next = fronthead;     /* add to front link list */
				fronthead = seg_p;
			}
			else   /* we need to split the line into two segments as the split line cross the seg line */
			{
				/* split the seg */
				*newseg_p = *seg_p;
				seg_p->max = splitpoint;
				newseg_p->min = splitpoint;

				seg_p->next = backhead;
				backhead = seg_p;
				newseg_p->next = fronthead;
				fronthead = newseg_p;
				newseg_p++;
			}
		}
	}

	*frontlist = fronthead;
	*backlist = backhead;
}

/*
================
=
= DrawMap
=
================
*/

static void DrawMap(void)
{
	int x, y;

	for (y = 0; y < 64; y++)
	{
		for (x = 0; x < 64; x++)
		{
			DrawTile(x, y, back[y][x]);
		}
	}
}

/*
================
=
= DrawTile
=
================
*/
static void DrawTile(int x, int y, int color)
{
	unsigned int red, green, blue;
	SDL_Rect aRect;

	red = (color & 3) * 85;
	green = ((color & 12) >> 2) * 85;
	blue = ((color & 48) >> 4) * 85;

	aRect.x = x * BLOCKSCALE;
	aRect.y = y * BLOCKSCALE;
	aRect.w = BLOCKSCALE;
	aRect.h = BLOCKSCALE;

	SDL_SetRenderDrawColor(gMapRenderer, red, green, blue, 0xFF);
	SDL_RenderSetScale(gMapRenderer, 1.0f, 1.0f);
	SDL_RenderFillRect(gMapRenderer, &aRect);
}

/*
=====================
=
= DrawSeg
=
=====================
*/

static void DrawSeg(SDL_Renderer* renderer, segment_t* seg)
{
	int coordinate, min, max;

	coordinate = seg->coordinate / 256;
	min = seg->min / 256;
	max = seg->max / 256;

	SDL_RenderSetScale(renderer, SCALE, SCALE);

	if (seg->orientation)
	{
		SDL_RenderDrawLine(renderer, (coordinate * BLOCKSCALE) / SCALE, (min * BLOCKSCALE) / SCALE, (coordinate * BLOCKSCALE) / SCALE, (max * BLOCKSCALE) / SCALE);
	}
	else
	{
		SDL_RenderDrawLine(renderer, (min * BLOCKSCALE) / SCALE, (coordinate * BLOCKSCALE) / SCALE, (max * BLOCKSCALE) / SCALE, (coordinate * BLOCKSCALE) / SCALE);
	}
}

/*
=====================
=
= DrawCircle
=
=====================
*/
static void DrawCircle(int32_t centreX, int32_t centreY, int32_t radius)
{
	const int32_t diameter = (radius * 2);

	int32_t x = (radius - 1);
	int32_t y = 0;
	int32_t tx = 1;
	int32_t ty = 1;
	int32_t error = (tx - diameter);

	while (x >= y)
	{
		/*  Each of the following renders an octant of the circle */
		SDL_RenderDrawPoint(gMapRenderer, centreX + x, centreY - y);
		SDL_RenderDrawPoint(gMapRenderer, centreX + x, centreY + y);
		SDL_RenderDrawPoint(gMapRenderer, centreX - x, centreY - y);
		SDL_RenderDrawPoint(gMapRenderer, centreX - x, centreY + y);
		SDL_RenderDrawPoint(gMapRenderer, centreX + y, centreY - x);
		SDL_RenderDrawPoint(gMapRenderer, centreX + y, centreY + x);
		SDL_RenderDrawPoint(gMapRenderer, centreX - y, centreY - x);
		SDL_RenderDrawPoint(gMapRenderer, centreX - y, centreY + x);

		if (error <= 0)
		{
			++y;
			error += ty;
			ty += 2;
		}

		if (error > 0)
		{
			--x;
			tx += 2;
			error += (tx - diameter);
		}
	}
}

/*
=================
=
= FloodMap
=
= areamap starts out with all solid tile being >= 0x80, and all open tiles having the soundarea (0-62)
= door tiles should have area 63
= the open areas are set to 0x7f as map.tilemap is filled in with the enclosed area numbers
=
=================
*/

static void FloodMap(void)
{
	int	x, y;
	int	tile;
	int	areanum;

	// create the areamap by setting all solid tiles to 0x80+tile and all open tiles to areanum
	for (y = 0; y < 64; y++)
	{
		for (x = 0; x < 64; x++)
		{
			tile = back[y][x];

			if (tile < AREATILES)
			{
				areamap[y][x] = 0x80 + tile;     /* solid tile */
				map.tilemap[y][x] = 0x80 + tile;
			}
			else
			{
				areamap[y][x] = tile - AREATILES;    /* open tile */
				map.tilemap[y][x] = 0;
			}
		}
	}

	// find every contained space in the map
	areanum = 0;

	for (y = 0; y < 64; y++)
	{
		for (x = 0; x < 64; x++)
		{
			tile = areamap[y][x];

			if (tile == 63)
			{
				map.tilemap[y][x] = 63;	// door area
			}
			else if (tile < 63)
			{						// unflooded open area
				map.areasoundnum[areanum] = tile;
				FloodFill(x, y, areanum);
				areanum++;
			}
		}
	}

	DrawMap();
}

/*
=================
=
= FloodFill
=
=================
*/

static void FloodFill(int x, int y, int fill)
{
	int	old;

	old = areamap[y][x];
	areamap[y][x] = 0x7f;		// this spot has been filled
	map.tilemap[y][x] = fill;
	back[y][x] = 0;

	if (areamap[y][x - 1] == old)
		FloodFill(x - 1, y, fill);

	if (areamap[y][x + 1] == old)
		FloodFill(x + 1, y, fill);

	if (areamap[y - 1][x] == old)
		FloodFill(x, y - 1, fill);

	if (areamap[y + 1][x] == old)
		FloodFill(x, y + 1, fill);
}


/*
================
=
= CreateSegments
=
================
*/
static void CreateSegments(void)
{
	int		x, y;
	int		state;
	int		tile, backtile;
	int	    topsolid, bottomsolid;
	int		min;

	// create segments
	//
	// Doors are entered as wall 0x80 + 0x40 * side + doornum

	/* set the line colour to black */
	SDL_SetRenderDrawColor(gMapRenderer, 0x00, 0x00, 0x00, 0xFF);

	/* do horizontal segs */
	for (y = 1; y < 64; y++)
	{
		state = BOTH_WALLS_SOLID;

		for (x = 1; x < 64; x++)
		{
			tile = back[y][x];
			backtile = back[y - 1][x];

			/* find out if back tile is solid */
			if (backtile >= DOORTILE && backtile <= LASTDOORTILE)
			{
				topsolid = 8;
			}
			else
			{
				topsolid = backtile < AREATILES;   /* if less then area tile (floor tile) its solid ie 1 else 0 */
			}

			if (tile >= DOORTILE && tile <= LASTDOORTILE)
			{
				bottomsolid = 8;
			}
			else
			{
				bottomsolid = tile < AREATILES;    /* if less then area tile (floor tile) its solid ie 1 else 0 */
			}

			/* check for a change in wall state */
			if (((topsolid << 1) + bottomsolid) == state)
			{
				continue;		// same as last tile
			}

			if (state == BOTTOM_WALL_SOLID)
			{
				AddSeg(di_west, (float)y, min, x, y);  // end of bottom solid wall
			}

			if (state == TOP_WALL_SOLID)
			{
				AddSeg(di_east, (float)y, min, x, y - 1);  // end of top solid wall
			}

			if (bottomsolid == 8)
			{	// handle door tiles explicitly
				if (topsolid)
				{	// vertical door
					AddSeg(di_east, (float)y, x, x + 1, 128);
					AddSeg(di_west, (float)(y + 1), x, x + 1, 128);
				}
				else
				{	// horizontal door
					AddSeg(di_east, float(y + 0.5f), x, x + 1, 129 + doornum[y][x]);
					AddSeg(di_west, float(y + 0.5f), x, x + 1, 129 + doornum[y][x]);
				}
			}

			state = (topsolid << 1) + bottomsolid;
			min = x;
		}
	}

	/* do vertical segs */
	for (x = 1; x < 64; x++)
	{
		state = BOTH_WALLS_SOLID;

		for (y = 1; y < 64; y++)
		{
			tile = back[y][x];
			backtile = back[y][x - 1];
	
			if (backtile >= DOORTILE && backtile <= LASTDOORTILE)
			{
				topsolid = 8;
			}
			else
			{
				topsolid = backtile < AREATILES;
			}

			if (tile >= DOORTILE && tile <= LASTDOORTILE)
			{
				bottomsolid = 8;
			}
			else
			{
				bottomsolid = tile < AREATILES;
			}

			if ((topsolid << 1) + bottomsolid == state)
			{
				continue;		// same as last tile
			}

			if (state == BOTTOM_WALL_SOLID)
			{
				AddSeg(di_south, (float)x, min, y, 64 + x);  // end of bottom solid wall
			}

			if (state == TOP_WALL_SOLID)
			{
				AddSeg(di_north, (float)x, min, y, 64 + x - 1);  // end of top solid wall
			}

			if (bottomsolid == 8)
			{	// handle door tiles explicitly
				if (topsolid)
				{	// horizontal door
					AddSeg(di_north, (float)x, y, y + 1, 128);
					AddSeg(di_south, (float)(x + 1), y, y + 1, 128);
				}
				else
				{	// vetical door
					AddSeg(di_south, (float)(x + 0.5f), y, y + 1, 129 + doornum[y][x]);
					AddSeg(di_north, (float)(x + 0.5f), y, y + 1, 129 + doornum[y][x]);
				}
			}

			state = (topsolid << 1) + bottomsolid;
			min = y;
		}
	}
}

/*
===============
=
= FindDoors
=
===============
*/
static void FindDoors(void)
{
	int x, y, tile;

	/* extract doors */
	numdoors = 0;
	memset(doornum, 0, sizeof(doornum));

	for (y = 0; y < 64; y++)
	{
		for (x = 0; x < 64; x++)
		{
			tile = back[y][x];

			if (tile < DOORTILE || tile > LASTDOORTILE)
			{
				continue;
			}

			front[y][x] = tile;
			doornum[y][x] = numdoors;
			numdoors++;
		}
	}
}

/*
===============
=
= RemovePWalls
=
===============
*/

static void RemovePWalls(void)
{
	int	x, y, area;

	numpwalls = 0;

	/* draw push wall segs in white */
	SDL_SetRenderDrawColor(gMapRenderer, 0xff, 0xff, 0xff, 0xFF);

	for (y = 0; y < 64; y++)
		for (x = 0; x < 64; x++)
		{
			if (front[y][x] != PUSHWALL)
				continue;

			if (back[y][x - 1] >= AREATILES)
				area = back[y][x - 1];
			else if (back[y][x + 1] >= AREATILES)
				area = back[y][x + 1];
			else if (back[y - 1][x] >= AREATILES)
				area = back[y - 1][x];
			else if (back[y + 1][x] >= AREATILES)
				area = back[y + 1][x];
			else
				Error("Enclosed pushwall at %i,%i", x, y);

			pwalls[numpwalls].x = x;
			pwalls[numpwalls].y = y;
			pwalls[numpwalls].area = area;
			pwalls[numpwalls].tile = back[y][x];
			numpwalls++;

            /* move the push wall back two tiles to its open position */
			if (front[y][x - 1] == SLOTTILE)
			{	// pushing to the west
				AddSeg(di_north, (float)(x + 1), y, y + 1, 64 + x);
				AddSeg(di_north, (float)x, y, y + 1, 64 + x);
				back[y][x - 2] = back[y][x];
			}
			else if (front[y][x + 1] == SLOTTILE)
			{	// pushing to the east
				AddSeg(di_south, (float)x, y, y + 1, 64 + x);
				AddSeg(di_south, (float)(x + 1), y, y + 1, 64 + x);
				back[y][x + 2] = back[y][x];
			}
			else if (front[y - 1][x] == SLOTTILE)
			{	// pushing to the north
				AddSeg(di_east, (float)y, x, x + 1, y);
				AddSeg(di_east, (float)(y + 1), x, x + 1, y);
				back[y - 2][x] = back[y][x];
			}
			else if (front[y + 1][x] == SLOTTILE)
			{	// pushing to the south
				AddSeg(di_west, (float)y, x, x + 1, y);
				AddSeg(di_west, (float)(y + 1), x, x + 1, y);
				back[y + 2][x] = back[y][x];
			}
			else
				Error("pushwall at %i,%i doesn't have an adjacent slot tile", x, y);

			/* make the push wall tile a normal floor tile */
			back[y][x] = area;
		}
}

/*
================
=
= AddSeg
=
================
*/

static void AddSeg(dir_t dir, float coordinate, int min, int max, int texture)
{
	newseg_p->dir = dir;
	newseg_p->orientation = (orientation_t)(1 ^ (dir & 1));
	newseg_p->coordinate = (int)(coordinate * 256.0f);
	newseg_p->min = min * 256;
	newseg_p->max = max * 256;
	newseg_p->texture = texture;
	DrawSeg(gMapRenderer, newseg_p);
	newseg_p++;
}


/*
================
=
= init
=
================
*/
static bool init(void)
{
	//Initialization flag
	bool success = true;
	int x;
	int y;

	//Initialize SDL
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		printf("SDL could not initialize! SDL_Error: %s\n", SDL_GetError());
		success = false;
	}
	else
	{
		//Set texture filtering to linear
		if (!SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "1"))
		{
			printf("Warning: Linear texture filtering not enabled!");
		}

		//Create map window
		gMapWindow = SDL_CreateWindow("Wolf3d bsp", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_SHOWN);
		if (gMapWindow == NULL)
		{
			printf("Map window could not be created! SDL Error: %s\n", SDL_GetError());
			success = false;
		}
		else
		{
			//Create MAP renderer for window
			gMapRenderer = SDL_CreateRenderer(gMapWindow, -1, SDL_RENDERER_ACCELERATED);
			if (gMapRenderer == NULL)
			{
				printf("Map renderer could not be created! SDL Error: %s\n", SDL_GetError());
				success = false;
			}
			else
			{
				//Initialize renderer color
				SDL_SetRenderDrawColor(gMapRenderer, 0xFF, 0xFF, 0xFF, 0xFF);

				/* get the map windows x,y  */
				SDL_GetWindowPosition(gMapWindow, &x, &y);

				//Create seg window
				gSegWindow = SDL_CreateWindow("BSP line segs", x + SCREEN_WIDTH, y, SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_SHOWN);
				if (gSegWindow == NULL)
				{
					printf("Seg window could not be created! SDL Error: %s\n", SDL_GetError());
					success = false;
				}
				else
				{
					//Create seg renderer for window
					gSegRenderer = SDL_CreateRenderer(gSegWindow, -1, SDL_RENDERER_ACCELERATED);
					if (gSegRenderer == NULL)
					{
						printf("Seg renderer could not be created! SDL Error: %s\n", SDL_GetError());
						success = false;
					}
					else
					{
						//Initialize renderer color
						SDL_SetRenderDrawColor(gSegRenderer, 0x00, 0x00, 0x00, 0xFF);
						SDL_RaiseWindow(gMapWindow);
					}
				}

			}
		}
	}

	return success;
}

/*
================
=
= close
=
================
*/
static void close(void)
{
	//Destroy window	
	SDL_DestroyRenderer(gMapRenderer);
	SDL_DestroyWindow(gMapWindow);

	gMapWindow = NULL;
	gMapRenderer = NULL;

	//Quit SDL subsystems
	SDL_Quit();
}